@@ -44,12 +44,36 @@
 #include <gdal.h>
 
 QgsOgrLayerItem::QgsOgrLayerItem( QgsDataItem *parent,
-                                  const QString &name, const QString &path, const QString &uri, LayerType layerType, bool isSubLayer )
+                                  const QString &name,
+                                  const QString &path,
+                                  const QString &uri,
+                                  LayerType layerType,
+                                  const QString &driverName,
+                                  bool isSubLayer )
   : QgsLayerItem( parent, name, path, uri, layerType, QStringLiteral( "ogr" ) )
+  , mDriverName( driverName )
+  , mIsSubLayer( isSubLayer )
 {
   mIsSubLayer = isSubLayer;
   mToolTip = uri;
-  setState( Populated ); // children are not expected
+  const bool isIndex { QRegularExpression( R"(=idx_[^_]+_[^_]+.*$)" ).match( uri ).hasMatch() };
+  setState( ( driverName ==  QStringLiteral( "SQLite" ) && ! isIndex ) ? NotPopulated : Populated ); // children are accepted except for sqlite
+}
+
+
+QVector<QgsDataItem *> QgsOgrLayerItem::createChildren()
+{
+  QVector<QgsDataItem *> children;
+  // Geopackage is handled by QgsGeoPackageVectorLayerItem and QgsGeoPackageRasterLayerItem
+  // Proxy to spatialite provider data items because it implements the connections API
+  if ( mDriverName == QLatin1String( "SQLite" ) )
+  {
+    children.push_back( new QgsFieldsItem( this,
+                                           path() + QStringLiteral( "/columns/ " ),
+                                           QStringLiteral( R"(dbname="%1")" ).arg( parent()->path().replace( '"', QLatin1String( R"(\")" ) ) ),
+                                           QStringLiteral( "spatialite" ), QString(), name() ) );
+  }
+  return children;
 }
 
 
@@ -107,12 +131,10 @@ QList<QgsOgrDbLayerInfo *> QgsOgrLayerItem::subLayers( const QString &path, cons
   // Vector layers
   const QgsVectorLayer::LayerOptions layerOptions { QgsProject::instance()->transformContext() };
   QgsVectorLayer layer( path, QStringLiteral( "ogr_tmp" ), QStringLiteral( "ogr" ), layerOptions );
-  if ( ! layer.isValid( ) )
-  {
-    QgsDebugMsgLevel( QStringLiteral( "Layer is not a valid %1 Vector layer %2" ).arg( path ), 3 );
-  }
-  else
+  if ( layer.isValid( ) )
   {
+    QVariantMap oriParts = QgsOgrProviderMetadata().decodeUri( path );
+
     // Collect mixed-geom layers
     QMultiMap<int, QStringList> subLayersMap;
     QgsOgrProvider *ogrProvider = qobject_cast<QgsOgrProvider *>( layer.dataProvider() );
@@ -122,7 +144,7 @@ QList<QgsOgrDbLayerInfo *> QgsOgrLayerItem::subLayers( const QString &path, cons
     int prevIdx = -1;
     for ( const QString &descriptor : subLayersList )
     {
-      QStringList pieces = descriptor.split( QgsDataProvider::SUBLAYER_SEPARATOR );
+      QStringList pieces = descriptor.split( QgsDataProvider::sublayerSeparator() );
       int idx = pieces[0].toInt();
       subLayersMap.insert( idx, pieces );
       if ( pieces.count() >= 4 && idx != prevIdx )
@@ -157,7 +179,6 @@ QList<QgsOgrDbLayerInfo *> QgsOgrLayerItem::subLayers( const QString &path, cons
         // example URI for mixed-geoms geoms:    '/path/gdal_sample_v1.2_no_extensions.gpkg|layerid=7|geometrytype=Point'
         // example URI for mixed-geoms attr table:    '/path/gdal_sample_v1.2_no_extensions.gpkg|layername=MyLayer|layerid=7'
         // example URI for single geoms:    '/path/gdal_sample_v1.2_no_extensions.gpkg|layerid=6'
-        QString uri;
         if ( layerType != QgsLayerItem::LayerType::NoType )
         {
           if ( geometryType.contains( QStringLiteral( "Collection" ), Qt::CaseInsensitive ) )
@@ -166,28 +187,33 @@ QList<QgsOgrDbLayerInfo *> QgsOgrLayerItem::subLayers( const QString &path, cons
           }
           else
           {
+            QVariantMap parts( oriParts );
             if ( uniqueNames )
-              uri = QStringLiteral( "%1|layername=%2" ).arg( path, name );
+              parts.insert( QStringLiteral( "layerName" ), name );
             else
-              uri = QStringLiteral( "%1|layerid=%2" ).arg( path, layerId );
+              parts.insert( QStringLiteral( "layerId" ), layerId );
             if ( values.size() > 1 )
             {
-              uri += QStringLiteral( "|geometrytype=" ) + geometryType;
+              parts.insert( QStringLiteral( "geometryType" ), geometryType );
             }
+            QString uri = QgsOgrProviderMetadata().encodeUri( parts );
             QgsDebugMsgLevel( QStringLiteral( "Adding %1 Vector item %2 %3 %4" ).arg( driver, name, uri, geometryType ), 3 );
-            children.append( new QgsOgrDbLayerInfo( path, uri, name, geometryColumn, geometryType, layerType ) );
+            children.append( new QgsOgrDbLayerInfo( path, uri, name, geometryColumn, geometryType, layerType, driver ) );
           }
         }
         else
         {
           QgsDebugMsgLevel( QStringLiteral( "Layer type is not a supported %1 Vector layer %2" ).arg( driver, path ), 3 );
-          uri = QStringLiteral( "%1|layerid=%2|layername=%3" ).arg( path, layerId, name );
-          children.append( new QgsOgrDbLayerInfo( path, uri, name, geometryColumn, geometryType, QgsLayerItem::LayerType::TableLayer ) );
+          QVariantMap parts( oriParts );
+          parts.insert( QStringLiteral( "layerId" ), layerId );
+          parts.insert( QStringLiteral( "layerName" ), name );
+          QString uri = QgsOgrProviderMetadata().encodeUri( parts );
+          children.append( new QgsOgrDbLayerInfo( path, uri, name, geometryColumn, geometryType, QgsLayerItem::LayerType::TableLayer, driver ) );
         }
-        QgsDebugMsgLevel( QStringLiteral( "Adding %1 Vector item %2 %3 %4" ).arg( driver, name, uri, geometryType ), 3 );
       }
     }
   }
+
   // Raster layers
   QgsRasterLayer::LayerOptions options;
   options.loadDefaultStyle = false;
@@ -200,8 +226,8 @@ QList<QgsOgrDbLayerInfo *> QgsOgrLayerItem::subLayers( const QString &path, cons
       // Split on ':' since this is what comes out from the provider
       QStringList pieces = uri.split( ':' );
       QString name = pieces.value( pieces.length() - 1 );
-      QgsDebugMsgLevel( QStringLiteral( "Adding GeoPackage Raster item %1 %2 %3" ).arg( name, uri ), 3 );
-      children.append( new QgsOgrDbLayerInfo( path, uri, name, QString(), QStringLiteral( "Raster" ), QgsLayerItem::LayerType::Raster ) );
+      QgsDebugMsgLevel( QStringLiteral( "Adding GeoPackage Raster item %1 %2" ).arg( name, uri ), 3 );
+      children.append( new QgsOgrDbLayerInfo( path, uri, name, QString(), QStringLiteral( "Raster" ), QgsLayerItem::LayerType::Raster, driver ) );
     }
   }
   else if ( rlayer.isValid( ) )
@@ -216,7 +242,7 @@ QList<QgsOgrDbLayerInfo *> QgsOgrLayerItem::subLayers( const QString &path, cons
 
     if ( ! hDS )
     {
-      QgsDebugMsg( QStringLiteral( "GDALOpen error # %1 : %2 " ).arg( CPLGetLastErrorNo() ).arg( CPLGetLastErrorMsg() ) );
+      QgsDebugMsgLevel( QStringLiteral( "GDALOpen error # %1 : %2 " ).arg( CPLGetLastErrorNo() ).arg( CPLGetLastErrorMsg() ), 2 );
 
     }
     else
@@ -235,9 +261,27 @@ QList<QgsOgrDbLayerInfo *> QgsOgrLayerItem::subLayers( const QString &path, cons
       }
 
       QgsDebugMsgLevel( QStringLiteral( "Adding %1 Raster item %2 %3" ).arg( driver, name, path ), 3 );
-      children.append( new QgsOgrDbLayerInfo( path, uri, name, QString(), QStringLiteral( "Raster" ), QgsLayerItem::LayerType::Raster ) );
+      children.append( new QgsOgrDbLayerInfo( path, uri, name, QString(), QStringLiteral( "Raster" ), QgsLayerItem::LayerType::Raster, driver ) );
     }
   }
+
+  // There were problems in reading the file: throw
+  if ( ! layer.isValid() && ! rlayer.isValid() && children.isEmpty() )
+  {
+    QString errorMessage;
+    // If it is file based and the file exists, there might be a permission error, let's change
+    // the message to give the user a hint about this possibility.
+    if ( QFile::exists( path ) )
+    {
+      errorMessage = tr( "The file does not contain any layer or there was an error opening the file.\nCheck file and directory permissions on\n%1" ).arg( QDir::toNativeSeparators( path ) );
+    }
+    else
+    {
+      errorMessage = tr( "Layer is not valid (%1)" ).arg( path );
+    }
+    throw QgsOgrLayerNotValidException( errorMessage );
+  }
+
   return children;
 }
 
@@ -291,7 +335,7 @@ static QgsOgrLayerItem *dataItemForLayer( QgsDataItem *parentItem, QString name,
   {
     // we are in a collection
     name = QString::fromUtf8( OGR_FD_GetName( hDef ) );
-    QgsDebugMsg( "OGR layer name : " + name );
+    QgsDebugMsgLevel( "OGR layer name : " + name, 2 );
     if ( !uniqueNames )
     {
       layerUri += "|layerid=" + QString::number( layerId );
@@ -306,7 +350,7 @@ static QgsOgrLayerItem *dataItemForLayer( QgsDataItem *parentItem, QString name,
 
   QgsDebugMsgLevel( "OGR layer uri : " + layerUri, 2 );
 
-  return new QgsOgrLayerItem( parentItem, name, path, layerUri, layerType, isSubLayer );
+  return new QgsOgrLayerItem( parentItem, name, path, layerUri, layerType, driverName, isSubLayer );
 }
 
 // ----
@@ -327,8 +371,8 @@ QVector<QgsDataItem *> QgsOgrDataCollectionItem::createChildren()
   CSLDestroy( papszOptions );
 
   GDALDriverH hDriver = GDALGetDatasetDriver( hDataSource.get() );
-  QString driverName = QString::fromUtf8( GDALGetDriverShortName( hDriver ) );
-  if ( driverName == QStringLiteral( "SQLite" ) )
+  const QString driverName = QString::fromUtf8( GDALGetDriverShortName( hDriver ) );
+  if ( driverName == QLatin1String( "SQLite" ) )
   {
     skippedLayerNames = QgsSqliteUtils::systemTables();
   }
@@ -352,7 +396,7 @@ QVector<QgsDataItem *> QgsOgrDataCollectionItem::createChildren()
       uniqueNames = false;
       break;
     }
-    if ( ( driverName == QStringLiteral( "SQLite" ) && layerName.contains( QRegularExpression( QStringLiteral( "idx_.*_geometry($|_.*)" ) ) ) )
+    if ( ( driverName == QLatin1String( "SQLite" ) && layerName.contains( QRegularExpression( QStringLiteral( "idx_.*_geometry($|_.*)" ) ) ) )
          || skippedLayerNames.contains( layerName ) )
     {
       skippedLayers << i;
@@ -403,6 +447,64 @@ bool QgsOgrDataCollectionItem::createConnection( const QString &name, const QStr
   return saveConnection( path, ogrDriverName );
 }
 
+bool QgsOgrDataCollectionItem::hasDragEnabled() const
+{
+  return true;
+}
+
+QgsMimeDataUtils::Uri QgsOgrDataCollectionItem::mimeUri() const
+{
+  QgsMimeDataUtils::Uri u;
+  u.providerKey = QStringLiteral( "ogr" );
+  u.uri = path();
+  u.layerType = QStringLiteral( "vector" );
+  return u;
+}
+
+QgsAbstractDatabaseProviderConnection *QgsOgrDataCollectionItem::databaseConnection() const
+{
+
+  QgsAbstractDatabaseProviderConnection *conn { QgsDataCollectionItem::databaseConnection() };
+
+  // There is a chance that this is a spatialite file, but spatialite is not handled by OGR and
+  // it's not even in core.
+  if ( ! conn )
+  {
+
+    // test that file is valid with OGR
+    if ( OGRGetDriverCount() == 0 )
+    {
+      OGRRegisterAll();
+    }
+    // do not print errors, but write to debug
+    CPLPushErrorHandler( CPLQuietErrorHandler );
+    CPLErrorReset();
+    gdal::dataset_unique_ptr hDS( GDALOpenEx( path().toUtf8().constData(), GDAL_OF_VECTOR, nullptr, nullptr, nullptr ) );
+    CPLPopErrorHandler();
+
+    if ( ! hDS )
+    {
+      QgsDebugMsgLevel( QStringLiteral( "GDALOpen error # %1 : %2 on %3" ).arg( CPLGetLastErrorNo() ).arg( CPLGetLastErrorMsg() ).arg( path() ), 2 );
+      return nullptr;
+    }
+
+    GDALDriverH hDriver = GDALGetDatasetDriver( hDS.get() );
+    QString driverName = GDALGetDriverShortName( hDriver );
+
+    if ( driverName == QLatin1String( "SQLite" ) )
+    {
+      QgsProviderMetadata *md { QgsProviderRegistry::instance()->providerMetadata( QStringLiteral( "spatialite" ) ) };
+      if ( md )
+      {
+        QgsDataSourceUri uri;
+        uri.setDatabase( path( ) );
+        conn = static_cast<QgsAbstractDatabaseProviderConnection *>( md->createConnection( uri.uri(), {} ) );
+      }
+    }
+  }
+  return conn;
+}
+
 // ---------------------------------------------------------------------------
 
 QString QgsOgrDataItemProvider::name()
@@ -410,6 +512,11 @@ QString QgsOgrDataItemProvider::name()
   return QStringLiteral( "OGR" );
 }
 
+QString QgsOgrDataItemProvider::dataProviderKey() const
+{
+  return QStringLiteral( "ogr" );
+}
+
 int QgsOgrDataItemProvider::capabilities() const
 {
   return QgsDataProvider::File | QgsDataProvider::Dir | QgsDataProvider::Net;
@@ -561,10 +668,17 @@ QgsDataItem *QgsOgrDataItemProvider::createDataItem( const QString &pathIn, QgsD
       QStringLiteral( "kml" ),
       QStringLiteral( "osm" ),
       QStringLiteral( "mdb" ),
+      QStringLiteral( "accdb" ),
+      QStringLiteral( "xls" ),
+      QStringLiteral( "xlsx" ),
+      QStringLiteral( "gpx" ),
+      QStringLiteral( "pdf" ),
       QStringLiteral( "pbf" ) };
   static QStringList sOgrSupportedDbDriverNames { QStringLiteral( "GPKG" ),
       QStringLiteral( "db" ),
       QStringLiteral( "gdb" ),
+      QStringLiteral( "xlsx" ),
+      QStringLiteral( "xls" ),
       QStringLiteral( "pgdb" )};
 
   // these extensions are trivial to read, so there's no need to rely on
@@ -632,7 +746,7 @@ QgsDataItem *QgsOgrDataItemProvider::createDataItem( const QString &pathIn, QgsD
 
   if ( ! hDS )
   {
-    QgsDebugMsg( QStringLiteral( "GDALOpen error # %1 : %2 on %3" ).arg( CPLGetLastErrorNo() ).arg( CPLGetLastErrorMsg() ).arg( path ) );
+    QgsDebugMsgLevel( QStringLiteral( "GDALOpen error # %1 : %2 on %3" ).arg( CPLGetLastErrorNo() ).arg( CPLGetLastErrorMsg() ).arg( path ), 2 );
     return nullptr;
   }
 
